#include "stdafx.h"
#include "DftIdftApp.h"

// constructor 
DftIdftApp::DftIdftApp(const int w, const int h)
{
	init(w, h);
}

void DftIdftApp::init(const int w, const int h)
{
	// O[XP[̉摜쐬
	src_img.create(h, w, CV_64F);
	cv::randn(src_img, 0.0, 1.0);

	Size s_size = src_img.size();
	src_cols = s_size.width;
	src_rows = s_size.height;

	// DFTÕf[^̏
	Ref_img = Mat(s_size, CV_64F);
	Imf_img = Mat::zeros(s_size, CV_64F);
	ReF_img = Mat(s_size, CV_64F);
	ImF_img = Mat::zeros(s_size, CV_64F);
	Complex_img = Mat(s_size, CV_64FC2);

	src_img.convertTo(Ref_img, CV_64F);
	mvf.push_back(Ref_img);
	mvf.push_back(Imf_img);
	mvF.push_back(ReF_img);
	mvF.push_back(ImF_img);
	merge(mvf, Complex_img);
	idft_img = zero = Mat::zeros(s_size, CV_64F);

	// dft̃TCYݒ
	dft_rows = getOptimalDFTSize(src_rows);
	dft_cols = getOptimalDFTSize(src_cols);
	dft_src = Mat::zeros(dft_rows, dft_cols, CV_64FC2);
	dft_dst = Mat::zeros(dft_rows, dft_cols, CV_64FC2);
	dft_dst_p = Mat::zeros(dft_dst.size(), dft_dst.type());
}

// DFTO/̎/f[^̐ݒ
void DftIdftApp::setValue(const bool isBefore, const int row, const int col, const double re, const double im)
{
	if (isBefore)
	{
		mvf[0].at<double>(row, col) = re;
		mvf[1].at<double>(row, col) = im;
	}
	else
	{
		mvF[0].at<double>(row, col) = re;
		mvF[1].at<double>(row, col) = im;
	}
}

// DFTO/̎/f[^̎擾
void DftIdftApp::getValue(const bool isBefore, const int row, const int col, double *re, double *im)
{
	if (isBefore)
	{
		*re = mvf[0].at<double>(row, col);
		*im = mvf[1].at<double>(row, col);
	}
	else
	{
		*re = mvF[0].at<double>(row, col);
		*im = mvF[1].at<double>(row, col);
	}
}

void DftIdftApp::calcDFT()
{
	// mvf}[W
	merge(mvf, Complex_img);

	// dft_srcɃf[^ݒ
	Mat roi(dft_src, Rect(0, 0, src_cols, src_rows));
	Complex_img.copyTo(roi);

	// dftsdft_dstɌʂi[
	dft(dft_src, dft_dst);
	dft_dst_p = Mat::zeros(dft_dst.size(), dft_dst.type());

	// Ƌɕ
	split(dft_dst, mvF);
}

// Inverse Discrete Fourier Transforma 
void DftIdftApp::calcIDFT(bool all)
{
	// ͏true
	if (all) 
	{
		merge(mvF, dft_dst);
		dft_dst.copyTo(dft_dst_p);
	}

	// tt[Gϊ(XP[O)
	idft(dft_dst_p, dft_src, DFT_SCALE);
	// 
	split(dft_src, mvf);
}

// s͗l̏(}bv)
void DftIdftApp::invertHeight()
{
	for (int i = 0; i<src_rows; i++)
	{
		for (int j = 1 - (i % 2); j<src_cols; j += 2)
		{
			mvf[0].at<double>(i, j) *= -1.0;
		}
	}
}

// s͗l̏(Choppy}bv)
void DftIdftApp::invertChoppy(const double chop)
{
	// MNtH
	for (int i = 0; i < src_rows; i++)
	{
		for (int j = 0; j < src_cols; j++)
		{
			if ((i + j) % 2 == 0)     // NOTE -- should be one!!! don't know why we lost a minus sign
			{
				mvf[0].at<double>(i, j) *= -1;    // cos(pi * (m1 + m2))
			}
			mvf[0].at<double>(i, j) *= chop;
		}
	}
}

// s͗l̏(m[}}bv)
void DftIdftApp::invertNormal(float *maxVal)
{
	float temp=0;
	for (int i = 0; i < src_rows; i++)
	{
		for (int j = 1 - (i % 2); j < src_cols; j += 2)
		{
			mvf[0].at<double>(i, j) *= -1;
			temp = fabs(mvf[0].at<double>(i, j));
			if (temp > *maxVal) *maxVal = temp;
		}
	}
}

void DftIdftApp::saveImage(const char *filename)
{
	double min, max;
	// ŏő擾
	minMaxLoc(mvf[0], &min, &max);
	// 摜𐶐
	idft_img = Mat(mvf[0] * 255.0 / max, Rect(0, 0, src_cols, src_rows));
	// 0`255ɐK
	normalize(idft_img, idft_img, 0, 255, CV_MINMAX);
	// ۑ
	cv::imwrite(filename, idft_img);
}
