#include "stdafx.h"
#include "OceanTest.h"
#include <cmath>
#include <iostream>
#include <vector>

#ifndef M_PI
#define M_PI 3.141592
#endif

// 1z}N
#define RELEASEBUFFER1D(p) if(p != NULL) { delete [] p; p = NULL; }

// 3xNg̐K
float normalize(float vec[3])
{
	float mag2 = vec[0]*vec[0]+vec[1]*vec[1]+vec[2]*vec[2];
	if (mag2 > FLT_EPSILON)
	{
		float mag = sqrt(mag2);
		float mul = 1.0f / mag;
		vec[0] *= mul;
		vec[1] *= mul;
		vec[2] *= mul;
		return mag;
	}
	return 0;
}

OceanTest::OceanTest()
{
	// e평
	Lx = Lz = 1000;
	M = N = 512;
	wind[0] = 1.0;
	wind[1] = 0.0;
	wind[2] = 0.0;
	V = 31.0;
	A = 1.0;
	chop = (0.5*Lx / (float)N);
	omega = NULL;
	h0ReF = NULL;
	h0ImF = NULL;
	position = NULL;
	normal = NULL;
	indices = NULL;
}


OceanTest::~OceanTest()
{
	// z̉
	release();
}

// 
void OceanTest::init(const int lx, const int lz, const int m, const int n, const float v, const float a, const float w[2])
{
	release();

	// p[^ݒ
	Lx = lx;
	Lz = lz;
	M = m;
	N = n;
	V = v;
	A = a;
	chop = (0.5*Lx / (float)N);
	wind[0] = w[0];
	wind[1] = w[1];
	wind[2] = 0.0;
	normalize(wind);

	// gvẐ߂̏
	int h0size = (M + 1)*(N + 1);
	omega = new float[h0size];
	h0ReF = new float[h0size];
	h0ImF = new float[h0size];
	calculateBaseHeight();

	// IDFT̏
	int hsize = M*N;
	height.init(M, N);
	normalDft[0].init(M, N);
	normalDft[1].init(M, N);
	choppy[0].init(M, N);
	choppy[1].init(M, N);

	// bV
	vertexNum = M*N;
	faceNum = (M - 1)*(N - 1) * 2;
	indexNum = 3 * faceNum;
	position = new float[3 * vertexNum];
	indices = new unsigned int[indexNum];
	normal = new float[M*N * 3];
}

// z̉
void OceanTest::release()
{
	RELEASEBUFFER1D(omega);
	RELEASEBUFFER1D(h0ReF);
	RELEASEBUFFER1D(h0ImF);
	RELEASEBUFFER1D(position);
	RELEASEBUFFER1D(indices);
	RELEASEBUFFER1D(normal);
}

// gvẐ߂̏
void OceanTest::calculateBaseHeight()
{
	// 
	cv::RNG rng;

	double multiplier, phk, y1, y2, y3, val1, val2;
	double k[2];

	for (int i = 0; i <= N; i++)
	{
		k[1] = (-N / 2.0f + i)*(2 * M_PI / Lz);
		for (int j = 0; j <= M; j++)
		{
			k[0] = (-M / 2.0f + j)*(2 * M_PI / Lx);
			double n = sqrt(k[0] * k[0] + k[1] * k[1]);
			omega[i*(M + 1) + j] = sqrt(9.81*n);

			if (k[0] == 0 && k[1] == 0)
			{
				multiplier = 0;
			}
			else
			{
				phk = PhillipsSpectrum(k);
				multiplier = sqrt(phk);
			}
			// K
			y1 = rng.gaussian(1.0);
			y2 = rng.gaussian(1.0);
			// l~2
			y3 = rng.uniform(0.0, 1.0)*2.0*M_PI;
			val1 = multiplier * y1 * sin(y3);
			val2 = multiplier * y2 * sin(y3);
			// ̎ƋiΏ́j
			h0ReF[i*(M + 1) + j] = val1;
			h0ImF[i*(M + 1) + j] = val2;
		}
	}
}

// Phillips Spectrum̌vZ
double OceanTest::PhillipsSpectrum(const double k[2])
{
	double L = V*V / 9.81;
	double l = Lx / 1000.0;

	double ksqr = k[0] * k[0] + k[1] * k[1];
	double kdotwhat = (double)(k[0] * wind[0] + k[1] * wind[1]);
	double phillips = (double)(A*exp(-1.0 / (L*L*ksqr)) / (ksqr*ksqr*ksqr)*(kdotwhat*kdotwhat));
	double damp = (double)exp(-ksqr*l*l);
	if (kdotwhat < 0) phillips *= 0.25f;
	return (phillips * damp);
}

// AChoppy, @xNǧvZ
void OceanTest::calculateHeight(const double t, const int normalMethod, const float yScale)
{
	float k[2];
	int mi, mj, index[2];
	double plus[2], minus[2];

	// ̌vZ
	for (int i = 0; i<N; i++)
	{
		k[1] = (-N / 2.0 + i)*(2.0*M_PI / Lz);
		mi = N - i;
		for (int j = 0; j<M; j++)
		{
			k[0] = (-M / 2.0 + j)*(2.0*M_PI / Lx);
			mj = M - j;
			index[0] = i*(M + 1) + j;
			index[1] = mi*(M + 1) + mj;
			plus[0] = (h0ReF[index[0]] * cos(omega[index[0]] * t) - h0ImF[index[0]] * sin(omega[index[0]] * t));
			plus[1] = (h0ReF[index[0]] * sin(omega[index[0]] * t) + h0ImF[index[0]] * cos(omega[index[0]] * t));
			minus[0] = (h0ReF[index[1]] * cos(-omega[index[0]] * t) - (-h0ImF[index[1]])*sin(-omega[index[0]] * t));
			minus[1] = (h0ReF[index[1]] * sin(-omega[index[0]] * t) + (-h0ImF[index[1]])*cos(-omega[index[0]] * t));
			// Ƌɒlݒ
			height.setValue(false, i, j, plus[0] + minus[0], plus[1] + minus[1]);
		}
	}
	// idft
	height.calcIDFT(true);
	height.invertHeight();

	// Choppy wavešvZ
	calculateChoppyWave();

	// @xNǧvZ
	if (normalMethod == 0)
	{
		// g@
		normalsFiniteDifferences(yScale);
	}
	else
	{
		// IDFTgp@iȂHj
		calculateNormals(yScale);
	}
}

// Choppy wavešvZ
void OceanTest::calculateChoppyWave()
{
	int i, j;
	float k[2], n;
	double re, im;
	// ijtHM=NȊȌꍇ̓G[o邩
	for (i = 0; i < M; i++)
	{
		k[0] = (float)(i - M / 2);
		for (j = 0; j < N; j++)         // -i (n1,n2)/sqrt(n1^2+n2^2) * htilda(n1, n2)
		{
			k[1] = (float)(j - N / 2);
			n = (float)sqrt(k[0] * k[0] + k[1] * k[1]);    // note n != 0
			if (n < FLT_EPSILON)
			{
				choppy[0].setValue(false, j, i, 0, 0);
				choppy[1].setValue(false, j, i, 0, 0);
			}
			else
			{
				height.getValue(false, j, i, &re, &im);
				choppy[0].setValue(false, j, i, im * k[0] / n, -re * k[0] / n);
				choppy[1].setValue(false, j, i, im * k[1] / n, -re * k[1] / n);
			}
		}
	}
	choppy[0].calcIDFT(true);
	choppy[0].invertChoppy(chop);
	choppy[1].calcIDFT(true);
	choppy[1].invertChoppy(chop);
}

/**
* This is a fairly cheap method for computing the normals, just taking finite differences.
* Currently requires $WX = WY$ (if not, need to divide out by the differences when setting
* nx and ny, and set nz to $1$).
*
* In my experience, this gives somewhat superior results to computing more accurate normals
* with the FFT unless small wavelength waves have been damped out.
*/
void OceanTest::normalsFiniteDifferences(const float yScale)
{
	int i, j, cnt = 1, index[9];
	float nx, ny, nz, nmag;

	for (i = 0; i < N; i++)
	{
		for (j = 0; j < N; j++, cnt += 3)
		{
			index[1] = cnt - 3 * N;
			index[0] = index[1] - 3;
			index[2] = index[1] + 3;
			index[4] = cnt;
			index[3] = cnt - 3;
			index[5] = cnt + 3;
			index[7] = cnt + 3 * N;
			index[6] = index[7] - 3;
			index[8] = index[7] + 3;

			for (int k = 0; k<9; k++)
			{
				if (index[k] < 0 || index[k] >= (3 * N*N)) index[k] = 0;
			}

			nx = -((2.0 * position[index[5]] + position[index[2]] + position[index[8]])
				- (2.0 * position[index[3]] + position[index[0]] + position[index[6]]));
			nz = -((2.0 * position[index[7]] + position[index[6]] + position[index[8]])
				- (2.0 * position[index[1]] + position[index[0]] + position[index[2]]));
			//     ny = 8 * Lx / (float) N;    // need WX = WY for this to work
			//		ny = 0.08;
			ny = yScale;
			//nx = -(px - mx);  // an even cheaper filter which gives similar results
			//ny = -(py - my);
			//nz = 2 * WX / (float) N;  // need WX = WY for this to work
			nmag = (float)sqrt(nx * nx + ny * ny + nz * nz); nx /= nmag; ny /= nmag; nz /= nmag;

			//      printf("n=%f,%f,%f\n", nx, ny, nz);
			normal[3 * (i*N + j) + 0] = nx;
			normal[3 * (i*N + j) + 1] = ny;
			normal[3 * (i*N + j) + 2] = nz;
		}
	}
}

// @xNǧvZQ
void OceanTest::calculateNormals(const float yScale)
{
	if (normal == NULL) return;
	float k[2], n[3];
	int i, j, index;
	float x, y, z, mag, maxnx = -1, maxnz = -1, tnx, tnz;
	index = 0;
	double re[3], im[3];
	for (i = 0; i < M; i++)
	{
		k[0] = (2 * M_PI / Lx) * (i - M / 2);
		for (j = 0; j < N; j++)
		{
			k[1] = (2 * M_PI / Lz) * (j - N / 2);
			height.getValue(false, i, j, &re[0], &im[0]);
			normalDft[0].setValue(false, i, j, -im[0] * k[0], re[0] * k[0]);
			normalDft[1].setValue(false, i, j, -im[0] * k[1], re[0] * k[1]);
			index++;
		}
	}
	normalDft[0].calcIDFT(true);
	normalDft[1].calcIDFT(true);
	normalDft[0].invertNormal(&maxnx);
	normalDft[1].invertNormal(&maxnz);

//	printf("max=%f,%f\n", maxnx, maxnz);
	index = 0;
	for (i = 0; i < M; i++)
	{
		for (j = 0; j < N; j++)
		{
			normalDft[0].getValue(true, i, j, &re[1], &im[1]);
			normalDft[1].getValue(true, i, j, &re[2], &im[2]);
			n[0] = (float)-re[1];
			//			n[1] = max(maxnz, maxnx);
			n[1] = yScale;
			n[2] = (float)-re[2];
			normalize(n);
			normal[3 * index + 0] = n[0];
			normal[3 * index + 1] = n[1];
			normal[3 * index + 2] = n[2];
			index++;
		}
	}
}

// bV̍XV
void OceanTest::updateHeight(const double t, const float scale[3], const float offset[3], const float chopScale, const int normalMethod, const float yScale)
{
	calculateHeight(t, normalMethod, yScale);
	int index;
	float dx = 2.0*scale[0] / (float)M;
	float dz = 2.0*scale[2] / (float)N;
	double re[3], im[3];
	for (int z = 0; z<N; z++)
	{
		for (int x = 0; x<M; x++)
		{
			index = z*M + x;
			height.getValue(true, z, x, &re[0], &im[0]);
			choppy[0].getValue(true, z, x, &re[1], &im[1]);
			choppy[1].getValue(true, z, x, &re[2], &im[2]);
			position[3 * index + 0] = offset[0] - scale[0] + dx * (float)x + chopScale*re[1];
			position[3 * index + 1] = offset[1] + scale[1] * re[0];
			position[3 * index + 2] = offset[2] - scale[2] + dz * (float)z + chopScale*re[2];
		}
	}
}

// bV̍쐬
void OceanTest::createMesh(const float scale[3], const float offset[3], const float chopScale)
{
	float dx = 2.0*scale[0] / (float)M;
	float dz = 2.0*scale[2] / (float)N;

	int index, cnt = 0;
	double re[3], im[3];
	for (int z = 0; z<N; z++)
	{
		for (int x = 0; x<M; x++)
		{
			index = z*M + x;
			height.getValue(true, z, x, &re[0], &im[0]);
			choppy[0].getValue(true, z, x, &re[1], &im[1]);
			choppy[1].getValue(true, z, x, &re[2], &im[2]);
//			printf("pos(%d,%d), h=%f, re=%f,%f\n", x, z, re[0], re[1], re[2]);
			position[3 * index + 0] = offset[0] - scale[0] + dx * (float)x + chopScale*re[1];
			position[3 * index + 1] = offset[1] + scale[1] * re[0];
			position[3 * index + 2] = offset[2] - scale[2] + dz * (float)z + chopScale*re[2];
			if (x < M - 1 && z < N - 1)
			{
				indices[cnt++] = index;
				indices[cnt++] = index + M;
				indices[cnt++] = index + 1;
				indices[cnt++] = index + 1;
				indices[cnt++] = index + M;
				indices[cnt++] = index + M + 1;
			}
		}
	}
}
